# -*- coding: utf-8 -*-

import os
from pathlib import Path
import yaml
from typing import List

from prefect.agent.local import LocalAgent

from simmate.configuration.prefect.connect_to_dask import set_default_executor
from simmate.configuration.dask.setup_cluster import run_cluster


def load_agent_settings() -> dict:
    """
    Load our default cluster type, agent name, and agent labels. This is loaded
    from the file `prefect_agent.yaml`.
    """

    SIMMATE_DIRECTORY = os.path.join(Path.home(), "simmate")
    AGENT_YAML = os.path.join(SIMMATE_DIRECTORY, "prefect_agent.yaml")
    if os.path.exists(AGENT_YAML):
        with open(AGENT_YAML) as file:
            settings = yaml.full_load(file)
    else:
        settings = {}

    # update each of these entries if it's not set in the yaml file
    default_values = {
        "agent_name": None,
        "agent_labels": [],
        "dask_cluster_type": "Local",
    }
    for key, default_value in default_values.items():
        if not settings.get(key):
            settings[key] = default_value

    return settings


DEFAULT_AGENT_SETTINGS = load_agent_settings()


def run_cluster_and_agent(
    cluster_type: str = DEFAULT_AGENT_SETTINGS["dask_cluster_type"],
    agent_name: str = DEFAULT_AGENT_SETTINGS["agent_name"],
    agent_labels: List[str] = DEFAULT_AGENT_SETTINGS["agent_labels"],
    **cluster_kwargs,
):

    # start our dask cluster
    cluster = run_cluster(
        cluster_type=cluster_type,
        **cluster_kwargs,
    )

    # We want Prefect to use our Dask Cluster to run all of the workflow tasks. To
    # tell Prefect to do this, we wrote a helper function (set_default_executor) that
    # ships with Simmate.
    set_default_executor(cluster.scheduler.address)

    agent = LocalAgent(
        name=agent_name,
        labels=agent_labels,
        hostname_label=False,
    )

    # Now we can start the Prefect Agent which will run and search for jobs and then
    # submit them to our Dask cluster.
    agent.start()
    # NOTE: this line will run endlessly unless you set a timelimit in the LocalAgent above



# -*- coding: utf-8 -*-

import math


class Incar(dict):
    """
    INCAR object for reading and writing INCAR files. This behaves exactly like
    a python dictionary, but has a few extra checks and methods attached to it.
    You can pass it a dictionary or initialize it just like you would dict(kwargs).
    You can consider the dict(kwargs) as equivalent to Incar(parameters).

    If you want a given setting to be dependent on the structure or dynamically
    determined, then we implement these modifiers. This would enable us to
    do things like ENCUT__per_atom or NGZF__density. We can even have more complex
    modifiers like multiple_keywords__smart_ldau which signals that our
    "smart_ldau" modifier introduces more than one new setting to the INCAR, such
    as LDAUJ, LDAUU, LDAUL, LDAUTYPE, and LDAUPRINT.

    TODO: In the future, I want to allow modifiers like __relative_to_previous
    and __use_previous to string settings accross tasks.

    """

    def __init__(self, **kwargs):

        # The kwargs are a dictionary of parameters (e.g. {"PREC": "accurate"})

        # Establish the dictionary with the given input
        # This is the same as doing dict(**kwargs)
        super().__init__(**kwargs)

        # Based on the kwargs, go through and make sure each parameter is in
        # it's expected datatype. Thus users can supply strings in the typical
        # VASP format that will be convert to python types here.
        # OPTIMIZE -- would it be faster if I only did this on from_file init?
        # and therefore assumed the user to initialize this with proper python
        # datatypes and formatting? Also will this behave properly if the value
        # is already in the correct format?
        for parameter, value in self.items():

            # parameters may have tags like "__density" added onto them. We don't
            # convert to the datatype yet, but instead wait until a structure
            # is provided (when the to_file or str methods are called below).
            # These are defined in python and should already be in the correct
            # data format.
            if "__" in parameter:
                self.update({parameter: value})

            # Otherwise, we need to convert values to the proper python datatype
            # because we might be reading from a file where everything is a string
            else:
                formatted_value = self._str_to_datatype(parameter, value)
                self.update({parameter: formatted_value})

        # SPECIAL CASE
        # If you have LSORBIT=True or LNONCOLLINEAR=True, the MAGMOM class needs
        # to take on a different form. Instead of a single MAGMOM being set as
        # a float, it's now a vector (x,y,z). For example:
        #   MAGMOM = x1 y1 z1    x2 y2 z2    x3 y3 z3 ...
        #       [...instead of...]
        #   MAGMOM = value1 value2 value3 ...
        # Check if MAGMOM is set along with LSORBIT or LNONCOLLINEAR being True
        if ("MAGMOM" in kwargs) and (
            kwargs.get("LSORBIT") or kwargs.get("LNONCOLLINEAR")
        ):
            # convert the MAGMOM value from...
            #   [x1,y1,z1,x2,y2,z2,x3,y3,z3] (this can be a string too)
            #   to...
            #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
            old_format = kwargs["MAGMOM"]
            new_format = [old_format[i : i + 3] for i in range(0, len(old_format), 3)]
            # Or If I wanted to do the same thing with numpy:
            #   new_format = numpy.array_split(old_format, len(old_format) // 3)
            # now update the dictionary with this value
            self.update({"MAGMOM": new_format})

    def to_evaluated_str(self, structure=None):

        # Let's start with an empty string and build from there
        final_str = ""

        # First we need to iterate through all parameters and check if we have
        # ones that are structure-specific. For example, we would need to
        # evaluate "ENCUT__per_atom". We go through all these and collect the
        # paramters into a final settings list.
        final_settings = {}
        for parameter, value in self.items():

            # if there is no modifier attached to the parameter, we just keep it as-is
            if "__" not in parameter:
                final_settings[parameter] = value

            # Otherwise we have a modifier like "__density" and need to evaluate it
            else:

                # make sure we have a structure supplied because all modifiers
                # require one.
                if not structure:
                    raise Exception(
                        "It looks like you used a keyword modifier but didn't "
                        f"supply a structure! If you want to use {parameter}, "
                        "then you need to make sure you provide a structure so "
                        "that the modifier can be evaluated."
                    )

                # separate the input into the base parameter and modifier.
                # This also overwrites what our paramter value is.
                parameter, modifier_tag = parameter.split("__")

                # check that this class has this modifier supported. It should
                # be a method named "keyword_modifier_mymodifer".
                # If everything looks good, we grab the modifier function
                modifier_fxn_name = "keyword_modifier_" + modifier_tag
                if hasattr(self, modifier_fxn_name):
                    modifier_fxn = getattr(self, modifier_fxn_name)
                else:
                    raise AttributeError(
                        """
                        It looks like you used a keyword modifier that hasn't
                        been defined yet! If you want something like ENCUT__smart_encut,
                        then you need to make sure there is a keyword_modifier_smart_encut
                        method available.
                        """
                    )

                # now that we have the modifier function, let's use it to update
                # our value for this keyword.
                value = modifier_fxn(structure, value)

                # if the "parameter" is actually "multiple_keywords", then we
                # have our actually parameters as a dictionary. We need to
                # pull these out of the "value" we have.
                if parameter == "multiple_keywords":
                    for subparameter, subvalue in value.items():
                        final_settings[subparameter] = subvalue

                # otherwise we were just given back an update value
                else:
                    final_settings[parameter] = value

        # Now that we have all of our parameters evaluated for the structure, we
        # iterate through each parameter and its set value. Each one will be
        # put on a separate line.
        for parameter, value in final_settings.items():

            # let's start by adding the parameter key to our output
            # It will be followed by an equal sign to separate it's value
            final_str += parameter + " = "

            # If we have a value that is a list (or list of lists)
            if isinstance(value, list):

                # check if we have a list of lists
                # Take MAGMOM with LSORBIT=True as an example, where we convert...
                #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
                # to...
                # "x1 y1 z1 x2 y2 z2 x3 y3 z3"
                if isinstance(value[0], list):
                    final_str += " ".join(str(n) for xyz in value for n in xyz)

                # otherwise we just have a list of items
                # converts [1, 2, 3] to "1 2 3"
                else:
                    final_str += " ".join(str(n) for n in value)

            # if it's not a list, we can just save the string of the value
            else:
                final_str += str(value)

            # we want to start each parameter on a new line, so we need to
            # add this at the end of each. There will be an extra new line at
            # the end of the file, but that's alright
            final_str += "\n"

        # we now have our final string and can return it!
        return final_str

    def to_file(self, filename="INCAR", structure=None):
        """
        Write Incar to a file.
        Args:
            filename (str): filename to write to.
        """
        # we just take the string format and put it in a file
        with open(filename, "w") as file:
            file.write(self.to_evaluated_str(structure=structure))

    @staticmethod
    def from_file(filename="INCAR"):
        """
        Reads an Incar object from a file.
        Args:
            filename (str): Filename for file
        Returns:
            Incar object
        """
        # open the file, grab the lines, and then close it
        with open(filename) as file:
            lines = file.readlines()

        # store parameters in this dictionary
        parameters = {}
        # go through line by line
        for line in lines:
            # If the line starts with a # then its a comment and we should skip.
            # It also could be an empty line that we should skip,
            if line.startswith("#") or line.startswith("\n"):
                continue
            # multiple parameters on a single line are separated by a semicolon
            for sub_line in line.split(";"):
                # the PARAMETER and VALUE are separated by equal sign
                parameter, value = sub_line.split("=")
                # we can use the parameter/value to update our dictionary. The
                # last thing we do is remove leading/trailing whitespace with strip()
                parameters[parameter.strip()] = value.strip()

        # return the final dictionary as an Incar object
        return Incar(**parameters)

    @staticmethod
    def _str_to_datatype(parameter, value):
        """
        When given a vasp parameter and it's value as a string, this helper
        function will use the key (parameter) to determine how to convert the
        val string to the proper python datatype (int, float, bool, list...).
        I have the most common keys mapped out, but if a parameter is given that
        isn't mapped, I simply leave it as a string.
        """

        # I outline the most common keys to what their expected data types are.

        # OPTIMIZE -- I should set these elsewhere so that these lists are not
        # initialized every time I call this function. Maybe have a dictionary
        # of {Parameter: Value_datatype} in the main enviornment for use.

        vector_list_keys = (
            # "MAGMOM",  # depends on other args -- see notes in init
            "DIPOL",
        )

        float_list_keys = (
            "LDAUU",
            "LDAUJ",
            "MAGMOM",  # depends on other args -- see notes in init
            "LANGEVIN_GAMMA",
            "QUAD_EFG",
            "EINT",
        )
        int_list_keys = (
            "LDAUL",
            "LDAUJ",
            "EINT",
        )

        bool_keys = (
            "LDAU",
            "LWAVE",
            "LSCALU",
            "LCHARG",
            "LPLANE",
            "LUSE_VDW",
            "LHFCALC",
            "ADDGRID",
            "LSORBIT",
            "LNONCOLLINEAR",
            "KGAMMA",
        )
        float_keys = (
            "EDIFF",
            "SIGMA",
            "TIME",
            "ENCUTFOCK",
            "HFSCREEN",
            "POTIM",
            "EDIFFG",
            "AGGAC",
            "PARAM1",
            "PARAM2",
            "KSPACING",
            "SYMPREC",
            "AMIX",
            "BMIX",
            "AMIN",
            "SMASS",
            "AMIX_MAG",
            "BMIX_MAG",
        )
        int_keys = (
            "NSW",
            "NBANDS",
            "NELMIN",
            "ISIF",
            "IBRION",
            "ISPIN",
            "ICHARG",
            "NELM",
            "ISMEAR",
            "NPAR",
            "LDAUPRINT",
            "LMAXMIX",
            "ENCUT",
            "NSIM",
            "NKRED",
            "NUPDOWN",
            "ISPIND",
            "LDAUTYPE",
            "IVDW",
            "ISTART",
            "NELMDL",
            "IMIX",
            "ISYM",
        )

        # If the value is not a string, then assume we are already in the
        # correct format. Note, an incorrect format will throw an error
        # somewhere below, which may be tricky for beginners to traceback.
        if not isinstance(value, str):
            return value

        # if the parameter is in int_keys
        if parameter in int_keys:
            # sometimes "1." was written to indicate an integer so check for
            # this and remove it if needed.
            if value[-1] == ".":
                value = value[:-1]
            # return the value integer
            return int(value)

        # if the parameter is in float_keys, we convert value to a float
        elif parameter in float_keys:
            # return the value float
            return float(value)

        # if the parameter is in bool_keys
        elif parameter in bool_keys:
            # Python is weird where bool("FALSE") will return True... So I need
            # to convert the string to lowercase and read it to know what to
            # return here.
            if "t" in value.lower():
                return True
            elif "f" in value.lower():
                return False

        # if the parameter is in vector_list_keys
        # These vectors are always floats
        elif parameter in vector_list_keys:
            # convert a string of...
            #   "x1 y1 z1 x2 y2 z2 x3 y3 z3"
            # to...
            #   [x1,y1,z1,x2,y2,z2,x3,y3,z3] (list of floats)
            # and then to...
            #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
            value = [float(item) for item in value.split()]
            return [value[i : i + 3] for i in range(0, len(value), 3)]

        # if the parameter is in float_list_keys
        elif parameter in float_list_keys:
            return [float(item) for item in value.split()]

        # if the parameter is in int_list_keys
        elif parameter in int_list_keys:
            return [int(item) for item in value.split()]

        # If it is not in the common keys listed, just leave it as a string.
        else:
            return value

    def compare_incars(self, other_incar):
        """
        Compares two Incars and indicates which parameters are the same and
        which are not. Useful for checking whether two runs were done using
        the same parameters.
        Args:
            other (Incar): The other Incar object to compare to.
        Returns:
            Dict of the following format:
            {"Same" : parameters_that_are_the_same,
            "Different": parameters_that_are_different}
            The value of the other_incar is returned for the paramters that
            are different are returned as {Parameter: (incar_value, other_incar_value)}
            where None is used as a placeholder. For the same parameters, the
            dictionary is returned as {Parameter: shared_value}
        """

        same_parameters = {}
        different_parameters = {}

        # make a collection of all unique keys used by the two dictionaries
        # I convert to list() so I can use set() below.
        # OPTIMIZE -- is there a better way to do this?
        parameters1 = list(self.keys())
        parameters2 = list(other_incar.keys())

        # iterate through each unique parameter
        for parameter in set(parameters1 + parameters2):

            # try to grab the value from both Incar objects
            # If it doesn't have the value, None will be provided
            value1 = self.get(parameter)
            value2 = other_incar.get(parameter)

            # compare the two and store
            if value1 == value2:
                # it doesn't matter which value I grab so I just use value1
                same_parameters[parameter] = value1
            # if they are different...
            else:
                different_parameters[parameter] = (value1, value2)

        return {"Same": same_parameters, "Different": different_parameters}

    def __add__(self, other):
        """
        Add all the values of another INCAR object to this object.
        Facilitates the use of "standard" INCARs.
        """
        # start by copying this incar's settings
        new_parameters = dict(self.items())

        # now iterate through the other incar parameters to incorporate them
        for parameter, value in other.items():

            # If both have a given parameter, make sure they are the same values
            if parameter in self and value != self[parameter]:
                # If not, we shouldn't allow combining these Incars as it can
                # lead to undesired results.
                raise ValueError(
                    f"Incars have conflicting values! One conflict is with {parameter} "
                    f"where one incar has {self[parameter]} and the other has {value}"
                )
            # otherwise just set the value
            else:
                new_parameters[parameter] = value

        # now return the new Incar
        return Incar(**new_parameters)

    # ------------------------------
    # All methods below are modifiers that users can apply. I may move these to
    # a separate class (e.g. KeywordModifier).

    @staticmethod
    def keyword_modifier_density(structure, density):
        """
        The __density modifier means the user wants a specific density. They
        provide this density in per-angstrom^3 units and we return the
        structure-specific count that gives this density.
        For example, density=10 and a structure lattice that volume of 5,
        then this returns value=10*5=50.
        """
        # VASP expect integers for a lot of these values, so we round up
        return math.ceil(structure.lattice.volume * density)

    @staticmethod
    def keyword_modifier_density_a(structure, density):
        """
        The __density_a modifier means the user wants a specific density along
        the A lattice vector. They provide this density in per-angstrom units
        and we return the structure-specific count that gives this density.
        For example, density=10 and a structure lattice that A vector of 5,
        then this returns value=10*5=50.
        """
        # VASP expect integers for a lot of these values, so we round up
        return math.ceil(structure.lattice.a * density)

    @staticmethod
    def keyword_modifier_density_b(structure, density):
        return math.ceil(structure.lattice.b * density)

    @staticmethod
    def keyword_modifier_density_c(structure, density):
        return math.ceil(structure.lattice.c * density)

    @staticmethod
    def keyword_modifier_per_atom(structure, per_atom_value):
        """
        The __per_atom modifier means the user wants a specific value per atom
        in the unit cell. For example, EDIFF__per_atom=1e-5 and a structure
        with 50 sites in it would return a value of 1e-5*50=50.
        """
        # VASP expect integers for a lot of these values, so we round up
        return per_atom_value * structure.num_sites

    @staticmethod
    def keyword_modifier_smart_magmom(structure, override_values):
        """
        The __smart_magmom modifier goes through a series of checks to decide
        what to set the MAGMOM as for VASP. In order of priority they are...
            (1) the magmom property attached to each site in the structure object
            (2) the spin property attached to the site's specie
            (3) a value provided explicitly (e.g. {"Co": 0.5})
            (4) a value of 0.6
        """

        # grab the default MAGMOM supplied, or use VASP's default of 1 otherwise
        default_value = override_values.get("default", 1)

        # we go through each site in the structure and decide what to set the
        # MAGMOM for each. This allows even different sites of the same
        # element to have their own MAGMOM
        magnetic_moments = []
        for site in structure:
            # if the structure object has magmom-decorated sites, we use that
            # as our first priority
            if hasattr(site, "magmom"):
                magnetic_moments.append(site.magmom)
            # next we check if the site's specie has a spin
            elif hasattr(site.specie, "spin"):
                magnetic_moments.append(site.specie.spin)
            # we then look at the override dictionary if there was one provided.
            # If note, we use 0.6 as a default.
            else:
                magnetic_moment = override_values.get(site.specie.symbol, default_value)
                magnetic_moments.append(magnetic_moment)

        # This feature is in pymatgen, but I haven't added it here yet.
        # if self.constrain_total_magmom:
        #     nupdown = sum([mag if abs(mag) > 0.6 else 0 for mag in incar["MAGMOM"]])
        #     incar["NUPDOWN"] = nupdown

        return magnetic_moments

    @staticmethod
    def keyword_modifier_smart_lmaxmix(structure, lmaxmix_config):
        """
        This modifier sets LMAXMIX if there are any d or f electrons present
        in the structure.
        """

        # COMMENT (from pymatgen team):
        # Note that if the user explicitly sets LMAXMIX in settings it will
        # override this logic.
        # Previously, this was only set if Hubbard U was enabled as per the
        # VASP manual but following an investigation it was determined that
        # this would lead to a significant difference between SCF -> NonSCF
        # even without Hubbard U enabled. Thanks to Andrew Rosen for
        # investigating and reporting.

        # first iterate through all elements and check for f-electrons
        if any(element.Z > 56 for element in structure.composition):
            return 6
        # now check for elements that contain d-electrons
        elif any(element.Z > 20 for element in structure.composition):
            return 4
        # otherwise use the default for VASP
        else:
            return 2

    @staticmethod
    def keyword_modifier_smart_ldau(structure, ldau_config):
        """
        This modifier handles a series of keyword arguments that are associated
        with LDAU, including LDAUJ, LDAUL, LDAUTYPE, LDAUU, LDAUPRINT, and LMAXMIX.
        Therefore, a complex dictionary is passed to this. The format looks like this...
            LDAU__multiple_keywords__smart_ldau = dict(
                LDAU__auto=True,
                LDAUTYPE=2,
                LDAUPRINT=1,
                LDAUJ={...},
                LDAUL={...},
                LDAUU={...},
                LMAXMIX__auto=True,
                )
        The LDAUJ, LDAUL, and LDAUU values can be a dictionary of elements to
        value, or (most commonly) a nested dictionary. For example...
            LDAUJ = {"F":{"Co":0}}
        This would mean if the structure is a fluoride, set the LDAUJ for Co to 0.
        If there are multiple options here (e.g. for fluorides and oxides), then
        priority is placed on the most electronegative element. This would mean
        some thing like yttrium oxide fluoride would be treated as a fluoride over
        an oxide.
        """

        # first we need to go through the LDAUJ, LDAUL, and LDAUU keywords and
        # see what their values are. If all of these end up be 0 for all elements
        # then we actually don't need LDAU at all! Therefore, we'll go through
        # all of these keywords and build a dictionary of settings to return
        # at the end of this function
        ldau_settings = {}

        # To help decide how we set these values, let's check what the most
        # electronegative element is, which will be last in the sorted composition
        most_electroneg = sorted(structure.composition.elements)[-1].symbol

        # As we go through these settings, we want see if we are even using LDAU.
        # For example if we ran a calculation on NaCl, we'd probably see that
        # all LDAUJ/LDAUL/LDAUU values are just 0. In that case, we can just
        # throw away (i.e. turn off) all LDAU settings. Therefore, we have a term
        # "using_ldau" that is false until proven otherwise.
        using_ldau = False

        for ldau_keyword in ["LDAUJ", "LDAUL", "LDAUU"]:

            # grab the sub-dictionary that maps elements to this keyword.
            # If it's not there, just use an empty dictionary.
            keyword_config = ldau_config.get(ldau_keyword, {})

            # check if the most electronegative element is in the override_values
            # and if so, see if it has a subdictionary in it. We use this
            # dictionary as our base one to pull values from.
            if most_electroneg in keyword_config and isinstance(
                keyword_config[most_electroneg], dict
            ):
                keyword_config = keyword_config[most_electroneg]

            # now iterate through all the sites and grab the assigned value. If nothing
            # is set then the default is 0.
            values = [
                keyword_config.get(element.symbol, 0)
                for element in structure.composition
            ]

            # now that we have this keyword all set, we add it to our results
            ldau_settings[ldau_keyword] = values

            # check to see if we are using ldau here (i.e. any value is not 0)
            if any(values):
                using_ldau = True

        # now check if we actaully need LDAU here. If not, we can throw out all
        # settings and just return an empty dictionary
        if using_ldau and "LDAU__auto" in ldau_config:
            ldau_settings["LDAU"] = True
        else:
            return {}

        # The remaining LDAU keywords are LDAUPRINT and LDAUTYPE, which we just
        # leave at what is set in the input
        if "LDAUPRINT" in ldau_config:
            ldau_settings["LDAUPRINT"] = ldau_config["LDAUPRINT"]
        if "LDAUTYPE" in ldau_config:
            ldau_settings["LDAUTYPE"] = ldau_config["LDAUTYPE"]

        return ldau_settings

    @staticmethod
    def keyword_modifier_smart_ismear(structure, ismear_config):
        """
        The smearing value used here depends on if we have a semiconductor,
        insulator, or metal. This modifier makes a "best-guess" on what the
        material is and uses the proper smearing type. Note that if this
        guess is wrong, it is useful to have the IncorrectSmearing error
        handler to fix this as VASP runs.

        Read more about the VASP recommended ISMEAR settings here:
            https://www.vasp.at/wiki/index.php/ISMEAR
        """

        # for now we just go through the structure and if all elements are
        # metals, then we say it's a metal. Otherwise, we treat the structure
        # as a semiconductor or insulator.
        if all(element.is_metal for element in structure.composition):
            ismear_settings = ismear_config.get("metal", {})

        else:
            ismear_settings = ismear_config.get("non-metal", {})

        return ismear_settings


# To introduce other modifiers that pymatgen uses...
# https://github.com/materialsproject/pymatgen/blob/b789d74639aa851d7e5ee427a765d9fd5a8d1079/pymatgen/io/vasp/sets.py#L500

# if self.use_structure_charge:
#     incar["NELECT"] = self.nelect

# Ensure adequate number of KPOINTS are present for the tetrahedron
# method (ISMEAR=-5). If KSPACING is in the INCAR file the number
# of kpoints is not known before calling VASP, but a warning is raised
# when the KSPACING value is > 0.5 (2 reciprocal Angstrom).
# An error handler in Custodian is available to
# correct overly large KSPACING values (small number of kpoints)
# if necessary.
# if "KSPACING" not in self.user_incar_settings.keys():
# if self.kpoints is not None:
#     if np.product(self.kpoints.kpts) < 4 and incar.get("ISMEAR", 0) == -5:
#         incar["ISMEAR"] = 0



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.static_energy.materials_project import (
    MatProjStaticEnergy,
)


class MatProjELF(MatProjStaticEnergy):
    """
    Runs a static energy calculation under settings from the Materials Project
    and also writes the electron localization function (to ELFCAR).
    """

    incar = MatProjStaticEnergy.incar.copy()
    incar.update(
        LELF=True,  # writes ELFCAR
        NPAR=1,  # must be set if LELF is set to True
        # BUG: if NPAR conflicts with INCAR_parallel_settings config this
        # fails and tells the user to specify a setting
    )



# -*- coding: utf-8 -*-

import django
from django.conf import settings

# BUG: import the settings raises an error in spyder. Should I moving this import
# inside of the setup_django_full() function?

# NOTE -- connecting just the Django database ORM has only a 0.05 second advantage
# on the initial import and then only 1.8e-7 second advantage after. Based on
# that, there really isn't much to gain from running the reduced setup. Even
# in ETL methods, I believe the bottleneck will still be the internet connection
# and not the django_setup. Having users always setup all of django can solve
# a lot of potential headaches at roughly the same speed.


def connect_database_only():  # Wall time: 200 ms first call and 600 ns after

    # see if django has already been configured. If so, just exit this function.
    if settings.configured:
        return

    # import the settings I want from the actual django settings file
    from simmate.configuration.django.settings import BASE_DIR, DATABASES, DEBUG

    # For speed, I only want this app installed
    # I also need to write out the full import path from django here.
    INSTALLED_APPS = ("simmate.website.diffusion.apps.DiffusionConfig",)
    # set these values

    settings.configure(
        BASE_DIR=BASE_DIR,
        DEBUG=DEBUG,
        DATABASES=DATABASES,
        INSTALLED_APPS=INSTALLED_APPS,
    )
    django.setup()


# upon import of this module, I run the setup immediately so the user doesn't have to
connect_database_only()



# -*- coding: utf-8 -*-

"""
Example use:
``` python
from simmate.toolkit import Structure
from simmate.workflows.static_energy import mit_workflow
from simmate.workflows.customized import vasp_workflow

s = Structure(
    lattice=[
        [3.485437, 0.0, 2.012318],
        [1.161812, 3.286101, 2.012318],
        [0.0, 0.0, 4.024635],
    ],
    species=["Na", "Cl"],
    coords=[[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]],
)

vasp_workflow.run(
    workflow_base=mit_workflow,
    input_parameters={"structure": s},
    updated_settings={
        "incar": {"NPAR": 1, "ENCUT": 600},
    },
)
```
"""

import prefect

from simmate.utilities import get_directory
from simmate.workflow_engine import Workflow, Parameter, ModuleStorage, task
from simmate.workflow_engine.common_tasks import run_customized_s3task, SaveOutputTask
from simmate.calculators.vasp.database.customized import CustomizedVASPCalculation


@task
def register_calc(workflow_base, input_parameters, updated_settings, source, directory):
    # !!! This code is copied and modified from the LoadInputAndRegister task.
    # I need to merge these functions in the future.

    directory_cleaned = get_directory(directory)
    input_parameters["directory"] = directory_cleaned

    prefect_flow_run_id = prefect.context.flow_run_id

    # HACK FIX: I need to make sure everything is JSON serailizable, and I use
    # a static method from the Workflow class to do this. This may be a good
    # sign that this method should become a utility instead.
    inputs_cleaned = Workflow._serialize_parameters(input_parameters)

    calculation = CustomizedVASPCalculation.from_prefect_id(
        id=prefect_flow_run_id,
        source=source or None,
        directory=directory_cleaned,
        workflow_base=workflow_base.name,
        input_parameters=inputs_cleaned,
        updated_settings=updated_settings,
    )

    return input_parameters


save_results = SaveOutputTask(CustomizedVASPCalculation)

with Workflow("customized/vasp") as vasp_workflow:

    workflow_base = Parameter("workflow_base")
    input_parameters = Parameter("input_parameters", default={})
    updated_settings = Parameter("updated_settings", default={})
    source = Parameter("source", default=None)
    directory = Parameter("directory", default=None)

    inputs_cleaned = register_calc(
        workflow_base=workflow_base,
        input_parameters=input_parameters,
        updated_settings=updated_settings,
        source=source,
        directory=directory,
    )

    result = run_customized_s3task(
        workflow_base=workflow_base,
        input_parameters=inputs_cleaned,
        updated_settings=updated_settings,
    )

    save_results(result)

vasp_workflow.storage = ModuleStorage(__name__)
vasp_workflow.project_name = "Simmate-Customized"
vasp_workflow.calculation_table = CustomizedVASPCalculation
vasp_workflow.result_table = CustomizedVASPCalculation
vasp_workflow.register_kwargs = {}
vasp_workflow.result_task = result
vasp_workflow.s3task = None
vasp_workflow.__doc__ = "VASP calculation with updated INCAR/POTCAR settings"



# -*- coding: utf-8 -*-

import os
import yaml

import prefect
from prefect import Task

from simmate.utilities import get_directory
from simmate.toolkit import Structure
from simmate.database.base_data_types import DatabaseTable

from typing import Tuple, Any

# OPTIMIZE: consider splitting this task into load_structure, load_directory,
# and register_calc so that our flow_visualize looks cleaner


class LoadInputAndRegister(Task):
    def __init__(
        self,
        workflow_name: str = None,
        input_obj_name: str = None,
        calculation_table: DatabaseTable = None,
        **kwargs,
    ):
        self.workflow_name = workflow_name
        self.input_obj_name = input_obj_name
        self.calculation_table = calculation_table  # If None, registration is skipped
        super().__init__(**kwargs)

    def run(
        self,
        input_obj: Any,
        input_class: Any = Structure,
        source: dict = None,
        directory: str = None,
        use_previous_directory: bool = False,
        **kwargs: Any,
    ) -> Tuple[Structure, str]:
        """
        How the input was submitted as a parameter depends on if we are submitting
        to Prefect Cloud, running the flow locally, or even continuing from a
        previous calculation.  Here, we use a task to convert the input to a toolkit
        object and (if requested) provide the directory as well.

        input_class is any class that has a from_dynamic method accepts the
        input_obj format. In the large majority of cases, this is just
        a Structure object so we use that as the default.

        directory is optional

        use_previous_directory is only used when we are pulling a structure from a
        previous calculation. If use_previous_directory=True, then the directory
        parameter is ignored.

        **kwargs is anything extra that you want saved to simmate_metadata.yaml
        """

        # -------------------------------------------------------------------------

        # First we load the structure to a toolkit structure object

        input_cleaned = input_class.from_dynamic(input_obj)

        # -------------------------------------------------------------------------

        # Now let's load the directory

        # if the user requested, we grab the previous directory as well
        if use_previous_directory and input_cleaned.is_from_past_calc:
            # this variable will only be set if the above conditions are met. In
            # this case we can grab the directory name for the simmate database entry
            directory_cleaned = input_cleaned.calculation.directory

        # catch incorrect use of this function
        elif use_previous_directory and not input_cleaned.is_from_past_calc:
            raise Exception(
                "There isn't a previous directory available! Your source structure "
                "must point to a past calculation to use this feature."
            )

        # otherwise use the directory that was given. We create this directly
        # immediately (rather than just passing the name to the S3Task). We
        # do this because NestedWorkflows often use a parent directory to
        # organize results.
        else:
            directory_cleaned = get_directory(directory)

        # This guards against incorrect use of the function too. We don't want
        # users asking to use a previous directory while also giving a brand
        # new directory.
        if use_previous_directory and directory:
            assert directory_cleaned == directory

        # -------------------------------------------------------------------------

        # Load the source of the input object

        # If we were given a input from a previous calculation, the source should
        # point directory to that same input. Otherwise we are incorrectly trying
        # to change what the source is.
        if source and input_cleaned.is_from_past_calc:
            # note input_obj here is a dictionary
            assert source == input_obj
        elif input_cleaned.is_from_past_calc:
            source_cleaned = input_obj
        elif source:
            source_cleaned = source
        else:
            source_cleaned = None

        # -------------------------------------------------------------------------

        # Register the calculation so the user can follow along in the UI.

        # This is only done if a table is provided. Some special-case workflows
        # don't store calculation information bc the flow is just a quick python
        # analysis.
        prefect_flow_run_id = prefect.context.flow_run_id
        if self.calculation_table:
            # load/create the calculation for this workflow run
            calculation = self.calculation_table.from_prefect_id(
                id=prefect_flow_run_id,
                # We pass the initial input_obj in case the calculation wasn't created
                # yet (and creation requires the structure)
                structure=input_cleaned,
                # BUG: what if the initial structure changed? An example of this happening
                # is with a relaxation where a correction was applied and the calc
                # was not fully restarted. This issue also will not matter when
                # workflows are ran through cloud -- as the structure is already
                # saved and won't be overwritten here.
                source=source_cleaned,
            )

        # -------------------------------------------------------------------------

        # Lastly, we want to write a file summarizing the inputs used for this
        # workflow run. This allows future users to reproduce the results if
        # desired -- and it also allows us to load old results into a database.
        input_summary = dict(
            workflow_name=self.workflow_name,
            source=source_cleaned,
            directory=directory_cleaned,
            # this ID is ingored as an input but needed for loading past data
            prefect_flow_run_id=prefect_flow_run_id,
            **kwargs,
        )
        # As a final thing to add, we want the input_obj, but we want to save
        # this to the proper input name. Also, this input should either be a
        # dictionary or a python object, where we convert to a dictionary in
        # order to write to file.
        input_dict = input_obj if isinstance(input_obj, dict) else input_obj.as_dict()
        input_summary[self.input_obj_name] = input_dict

        # now write the summary to file in the same directory as the calc.
        input_summary_filename = os.path.join(
            directory_cleaned, "simmate_metadata.yaml"
        )
        with open(input_summary_filename, "w") as file:
            content = yaml.dump(input_summary)
            file.write(content)

        # -------------------------------------------------------------------------

        # the rest of the calculation doesn't need the source (that was only for
        # registering the calc), so we just return the pymatgen structure and dir
        return input_cleaned, directory_cleaned



# -*- coding: utf-8 -*-

from simmate.workflow_engine import s3task_to_workflow
from simmate.calculators.vasp.tasks.static_energy import (
    NEBEndpointStaticEnergy as NEBEndpointStaticEnergyTask,
)
from simmate.calculators.vasp.database.energy import (
    NEBEndpointStaticEnergy as NEBEndpointStaticEnergyResults,
)

workflow = s3task_to_workflow(
    name="static-energy/neb-endpoint",
    module=__name__,
    project_name="Simmate-Energy",
    s3task=NEBEndpointStaticEnergyTask,
    calculation_table=NEBEndpointStaticEnergyResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)



# -*- coding: utf-8 -*-

import json
import cloudpickle

# note: extra modules are imported from prefect for convenience imports elsewhere
import prefect
from prefect import Client
from prefect import Flow as PrefectFlow, Parameter, task, Task
from prefect.storage import Module as ModuleStorage
from prefect.utilities.graphql import with_args
from prefect.backend import FlowRunView, FlowView

# This import below is used to grab `watch_flow_run`, but we don't grab it directly
# in order to support "mocking" this function in unit tests
from prefect.backend import flow_run as flow_run_module

from typing import List
from simmate.workflow_engine.supervised_staged_shell_task import S3Task
from simmate.database.base_data_types import DatabaseTable, Calculation


class Workflow(PrefectFlow):
    """
    This class behaves exactly like a normal Prefect workflow, where we add some
    common utilities and pre-submit tasks. For example, there is the `run_cloud`
    method, which allows us to register a calculation to a database table before
    we submit the workflow to Prefect Cloud.

    To learn how to use this class, see [prefect.core.flow.Flow](https://docs.prefect.io/api/latest/core/flow.html#flow-2)
    """

    result_task: Task = None
    """
    In many ETL workflows, we may also want the result of the some terminating
    task directly. This would save us from having to go find the result in
    the database. So by setting `result_task`, we are able get access a specific task's
    result -- directly as a python object(s). This is optional.
    """

    project_name: str = None
    """
    The name of the project in the  that this workflow is associated with. This 
    attribute is mainly just for organizing workflows in the Prefect Cloud interface.
    """

    s3task: S3Task = None
    """
    The supervised-staged-shell task (or S3Task) that this workflow uses to run.
    For understanding what the calculation does and the settings it uses, users
    should start here. You can also use a workflows `s3task.run` to run the workflow
    without storing results in the database.
    """

    calculation_table: Calculation = None
    """
    The database table where calculation information (such as the prefect_flow_run_id)
    is stored. Note, for NestedWorkflows, this table will not be the same as the
    result table! The table should use 
    `simmate.database.base_data_types.Calculation`
    """

    result_table: DatabaseTable = None
    """
    The database table where all calculation results are stored. In many cases,
    this is the same table as `calculation_table` -- the exception is
    for NestedWorkflows, where the result table may point to a specific
    sub-workflow's table for results. An example of this is the relaxation/staged
    workflow, which is made up of a series of relaxations -- and the result 
    table points to the final relaxation in this series. The table should use 
    `simmate.database.base_data_types.DatabaseTable`
    """

    @property
    def type(self) -> str:
        """
        Gives the workflow type of this workflow. For example the workflow named
        'static-energy/matproj' would have the type `static-energy`.
        """
        return self.name.split("/")[0]

    @property
    def name_short(self) -> str:
        """
        Gives the present name of the workflow. For example the workflow named
        'static-energy/matproj' would have the shortname `matproj`
        """
        return self.name.split("/")[-1]

    # BUG: naming this `description` causes issues.
    # See https://github.com/PrefectHQ/prefect/issues/3911
    @property
    def description_doc(self) -> str:
        """
        This simply returns the documentation string of this workflow -- so this
        is the same as `__doc__`. This attribute is only defined for beginners
        to python and for use in django templates for the website interface.
        """
        return self.__doc__

    def run_cloud(
        self,
        labels: List[str] = [],
        wait_for_run: bool = True,
        **kwargs,
    ) -> str:
        """
        This schedules the workflow to run remotely on Prefect Cloud.

        #### Parameters

        - `labels`:
            a list of labels to schedule the workflow with

        - `wait_for_run`:
            whether to wait for the workflow to finish. If False, the workflow
            will simply be submitted and then exit. The default is True.

        - `**kwargs`:
            all options that are normally passed to the workflow.run() method

        #### Returns

        - The flow run id that was used in prefect cloud.


        #### Usage

        Make sure you have Prefect properly configured and have registered your
        workflow with the backend.

        Note that this method can be viewed as a fork of:
            - from prefect.tasks.prefect.flow_run import create_flow_run
        It can also be viewed as a more convenient way to call to client.create_flow_run.
        I don't accept any other client.create_flow_run() inputs besides 'labels'.
        This may change in the future if I need to set flow run names or schedules.
        """

        # Grab the logger as we will print useful information below.
        # In some cases, there is no logger present, so we just set logger=None.
        # This possibility is why we use `if logger else None` statements below.
        logger = getattr(prefect.context, "logger", None)

        # Grab the Flow's ID from Prefect Cloud
        logger.debug("Looking up flow metadata...") if logger else None
        flow_view = FlowView.from_flow_name(
            self.name,
            project_name=self.project_name,
        )

        # Serialize all of the parameters so they can be properly submitted
        parameters_serialized = self._serialize_parameters(kwargs)

        # Now we submit the workflow
        logger.info(f"Creating flow run for {self.name}...") if logger else None
        client = Client()
        flow_run_id = client.create_flow_run(
            flow_id=flow_view.flow_id,
            parameters=parameters_serialized,
            labels=labels,
        )

        # we log the website url to the flow for the user
        run_url = client.get_cloud_url("flow-run", flow_run_id, as_user=False)
        logger.info(f"Created flow run: {run_url}") if logger else None

        # -----------------------------------
        # This part is unique to Simmate. Because we often want to save some info
        # to our database even before the calculation starts/finishes, we do that
        # here. An example is storing the structure and prefect id that we just
        # submitted.
        # BUG: Will there be a race condition here? What if the workflow finishes
        # and tries writing to the databse before this is done?
        register_kwargs = {
            key: kwargs[key] for key in kwargs if key in self.register_kwargs
        }
        calc = self._register_calculation(
            flow_run_id,
            **register_kwargs,
        )  # !!! should I return the calc to the user?
        # -----------------------------------

        # if we want to wait until the job is complete, we do that here
        if wait_for_run:
            flow_run_view = self.wait_for_flow_run(flow_run_id)
            # return flow_run_view # !!! Should I return this instead?
        # return the flow_run_id for the user
        return flow_run_id

    def _register_calculation(self, flow_run_id: str, **kwargs):
        """
        If the workflow is linked to a calculation table in the Simmate database,
        this adds the flow run to the Simmate database.

        This method should not be called directly as it is used within the
        run_cloud() method.
        """
        # If there's no calculation database table in Simmate for this workflow,
        # just skip this step. Otherwise save/load the calculation to our table
        if self.calculation_table:
            calculation = self.calculation_table.from_prefect_id(
                id=flow_run_id,
                **kwargs,
            )
            return calculation

    def wait_for_flow_run(self, flow_run_id: str):
        """
        Waits for a given flow run to complete

        This method is a direct fork of...
            from prefect.tasks.prefect.flow_run import wait_for_flow_run

        It does exactly the same thing where I just assume I want to stream logs.
        """

        flow_run = FlowRunView.from_flow_run_id(flow_run_id)

        for log in flow_run_module.watch_flow_run(flow_run_id):
            message = f"Flow {flow_run.name}: {log.message}"
            prefect.context.logger.log(log.level, message)
        # Return the final view of the flow run
        return flow_run.get_latest()

    @staticmethod
    def _serialize_parameters(parameters: dict) -> dict:
        """
        Converts input parameters to json-sealiziable objects that Prefect can
        use.

        This method should not be called directly as it is used within the
        run_cloud() method.
        """
        # Because many flows allow object-type inputs (such as structure object),
        # we need to serialize these inputs before scheduling them with prefect
        # cloud. To do this, I only check for two potential methods:
        #     as_dict
        #     to_dict
        # These were chosen based on common pymatgen methods, but I may change
        # this in the future. As another alternative, I could also cloudpickle
        # input objects when they are not JSON serializable.
        # OPTIMIZE: Prefect current tries to JSON serialize within their
        # client.create_flow_run method. In the future, we may want to move
        # this functionality there.

        parameters_serialized = {}
        for parameter_key, parameter_value in parameters.items():

            try:
                json.dumps(parameter_value)
            except TypeError:
                if hasattr(parameter_value, "as_dict"):
                    parameter_value = parameter_value.as_dict()
                elif hasattr(parameter_value, "to_dict"):
                    parameter_value = parameter_value.to_dict()
                else:
                    parameter_value = cloudpickle.dumps(parameter_value)
            parameters_serialized[parameter_key] = parameter_value
        return parameters_serialized

    @property
    def nflows_submitted(self) -> int:
        """
        Queries Prefect to see how many workflows are in a running or submitted
        state. It will return a count (integer).

        Note, your workflow must be registered with Prefect for this to work.
        """
        query = {
            "query": {
                with_args(
                    "flow_run_aggregate",
                    {
                        "where": {
                            "flow": {"name": {"_eq": self.name}},
                            "state": {"_in": ["Running", "Scheduled"]},
                        },
                    },
                ): {"aggregate": {"count"}}
            }
        }
        client = Client()
        result = client.graphql(query)
        return result["data"]["flow_run_aggregate"]["aggregate"]["count"]

    def to_workflow_task(self):
        """
        Converts a prefect workflow to a prefect task (aka a "workflow task")

        See the documentation in workflow_engine.tasks.workflow_task for more.
        """
        from simmate.workflow_engine import WorkflowTask

        return WorkflowTask(workflow=self)

    # @classmethod
    # def get_all_ids_in_state(cls, states=["Completed", "Scheduled"]):
    #     # TODO: This is an example query where I grab all ids. This may be useful in the
    #     # future.
    #     query = {
    #         "query": {
    #             with_args(
    #                 "flow_run",
    #                 {
    #                     "where": {
    #                         "flow": {"name": {"_eq": cls.name}},
    #                         "state": {"_in": states},
    #                     },
    #                 },
    #             ): ["id"]
    #         }
    #     }



# -*- coding: utf-8 -*-

# TODO: when I add more calculators, I can do something like this...
# if "simmate.calculators.vasp" in installed_apps:
from simmate.calculators.vasp.workflows.population_analysis import (
    bader_matproj_workflow,
    elf_matproj_workflow,
)



# -*- coding: utf-8 -*-

import os

import django
from django.conf import settings


def setup_full():
    """
    Configures django settings. This method does not need to be called directly
    because importing this module does so automatically.

    The first time this function is used can take >1 seconds because many things
    are imported from the start. Thereafter, testing shows that only <200 ns
    are used when settings are already configured.
    """

    # OPTIMIZE: this function will take longer than 250ms when other imports
    # get involved. For example, simmate.database.diffusion has some pymatgen
    # imports that that >0.8s and make this function take over 1s. I should have
    # setup_select() function that only configures "select" apps to help with
    # setup speed.

    # see if django has already been configured. If so, just exit this function.
    # BUG: originally I used the code below, but it didn't work with Prefect+Dask:
    #   if "DJANGO_SETTINGS_MODULE" in os.environ:
    if settings.configured:
        return

    # The code below is the equiv of running 'python manage.py shell'
    # This sets up all of django for its full use. You can being importing
    # models after this is setup.
    os.environ.setdefault(
        "DJANGO_SETTINGS_MODULE", "simmate.configuration.django.settings"
    )
    django.setup()

    # SECURITY WARNING: I added this setting myself! delete when in production
    # BUG: I believe this is just for use within Spyder and Jupyter, which I think
    # are examples of async enviornments. In production (non-dev), I believe I should
    # turn this off though.
    os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "true"


# When I import this module, it automatically configures django for us, including
# connecting to the database(s). Without this file, I would instead need these two
#  lines in every single file before I import models:
#   from simmate.configuration.django import setup_full
#   setup_full() # ensures setup
# With this init, I instead only need to have one line:
#   from simmate.configuration.django import setup_full # ensures setup
# Thus this "module" is more of a script.
setup_full()



# -*- coding: utf-8 -*-

from .materials_project import workflow as matproj_workflow
from .mit import workflow as mit_workflow
from .quality_04 import workflow as quality04_workflow
from .neb_endpoint import workflow as neb_endpoint_workflow



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow, Parameter, ModuleStorage
from simmate.workflow_engine.common_tasks import (
    LoadInputAndRegister,
    SaveOutputTask,
)

from typing import List
from simmate.workflow_engine.supervised_staged_shell_task import S3Task
from simmate.database.base_data_types import Calculation


def s3task_to_workflow(
    name: str,
    module: str,
    project_name: str,
    s3task: S3Task,
    calculation_table: Calculation,
    register_kwargs: List[str],
):
    """
    Builds a workflow from a S3Task and it's corresponding database table.

    Very often with Simmate's S3Tasks, the workflow for a single S3Task is
    the same. The workflow is typically made of three tasks:

    1. loading the input parameters and registering the calculation
    2. running the calculation (what this S3Task does on its own)
    3. saving the calculation results

    Task 1 and 3 always use the same functions, where we just need to tell
    it which database table we are registering/saving to.

    Because of this common recipe for workflows, we use this method to make
    the workflow for us.
    """

    s3task_obj = s3task()  # Use defaults
    load_input_and_register = LoadInputAndRegister(
        workflow_name=name,
        input_obj_name="structure",
        calculation_table=calculation_table,
    )
    save_results = SaveOutputTask(calculation_table)

    with Workflow(name) as workflow:
        structure = Parameter("structure")
        command = Parameter("command", default="vasp_std > vasp.out")
        source = Parameter("source", default=None)
        directory = Parameter("directory", default=None)
        use_previous_directory = Parameter("use_previous_directory", default=False)

        structure_toolkit, directory_cleaned = load_input_and_register(
            input_obj=structure,
            command=command,
            source=source,
            directory=directory,
            use_previous_directory=use_previous_directory,
        )
        output = s3task_obj(
            structure=structure_toolkit,
            command=command,
            directory=directory_cleaned,
        )
        calculation_id = save_results(output=output)

    workflow.storage = ModuleStorage(module)
    workflow.project_name = project_name
    workflow.calculation_table = calculation_table
    workflow.result_table = calculation_table
    workflow.register_kwargs = register_kwargs
    workflow.result_task = output
    workflow.s3task = s3task

    # by default we just copy the docstring of the S3task to the workflow
    workflow.__doc__ = s3task.__doc__

    return workflow



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.relaxation.materials_project import (
    MatProjRelaxation,
)

from simmate.calculators.vasp.error_handlers import (
    TetrahedronMesh,
    Eddrmm,
    NonConverging,
)


class MatProjStaticEnergy(MatProjRelaxation):
    """
    Runs a VASP static energy calculation using Materials Project settings.

    This is identical to relaxation/matproj, but just a single ionic step.
    """

    # The settings used for this calculation are based on the MITRelaxation, but
    # we are updating/adding new settings here.
    # !!! we hardcode temperatures and time steps here, but may take these as inputs
    # in the future
    incar = MatProjRelaxation.incar.copy()
    incar.update(
        dict(
            IBRION=-1,  # (optional) locks everything between ionic steps
            NSW=0,  # this is the main static energy setting
        )
    )

    # We reduce the number of error handlers used on static because many
    # error handlers are based on finding the global minimum & converging, which
    # not what we're doing with static
    error_handlers = [
        TetrahedronMesh(),
        Eddrmm(),
        NonConverging(),
    ]



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)
from simmate.calculators.vasp.error_handlers import (
    TetrahedronMesh,
    RotationMatrix,
    Brmix,
    Zpotrf,
    SubspaceMatrix,
    IncorrectShift,
    RealOptlay,
    Tetirr,
    RotationNonIntMatrix,
    LongVector,
    TripleProduct,
    Pricel,
    Brions,
    Zbrent,
    InsufficientBands,
    Pssyevx,
    Eddrmm,
    Edddav,
    Edwav,
    Zheev,
    ElfKpar,
    Rhosyg,
    Posmap,
    PointGroup,
    SymprecNoise,
    IncorrectSmearing,
    MeshSymmetry,
    Unconverged,
    NonConverging,
    Potim,
    PositiveEnergy,
    Frozen,
    LargeSigma,
)


class MITRelaxation(VaspTask):
    """
    Runs a VASP geometry optimization using MIT Project settings.

    The MIT Project is the precursor to the Materials Project.

    From their documention in pymatgen:

        "Default VASP settings for calculations in the pre-cursor MIT project to
        the Materials Project. Reasonably robust, but selected PSP are generally the
        ones with fewer electrons and convergence criteria is less tight. This was
        used in an era whether computational power is more limited.
        Nevertheless, this is still a good starting point for extremely expensive
        methods."

    """

    # This uses the PBE functional with POTCARs that have lower electron counts
    # and convergence criteria.
    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS_LOW_QUALITY

    # These are all input settings for this task. Note a lot of settings
    # depend on the structure/composition being analyzed.
    incar = dict(
        # These settings are the same for all structures regardless of composition
        ALGO="Fast",
        EDIFF=1.0e-05,
        ENCUT=520,
        IBRION=2,
        ICHARG=1,
        ISIF=3,
        ISPIN=2,
        ISYM=0,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=200,
        NELMIN=6,
        NSW=99,
        PREC="Accurate",
        KSPACING=0.5,  # !!! This is where we are different from pymatgen right now
        # The magnetic moments are dependent on what the composition and oxidation
        # states are. Note our default of 0.6 is different from the VASP default too.
        MAGMOM__smart_magmom={
            "default": 0.6,
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        # The type of smearing we use depends on if we have a metal, semiconductor,
        # or insulator. So we need to decide this using a keyword modifier.
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=2,
                SIGMA=0.2,
            ),
            "non-metal": dict(
                ISMEAR=-5,
                SIGMA=0.05,
            ),
        },
        # Changing limits of mixing for larger atoms
        LMAXMIX__smart_lmaxmix=True,
        # We run LDA+U for certain compositions. This is a complex configuration
        # so be sure to read the "__smart_ldau" modifier for more information.
        # But as an example for how the mappings work...
        # {"F": {"Ag": 2}} means if the structure is a fluoride, then we set
        # the MAGMOM for Ag to 2. Any element that isn't listed defaults to 0.
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LDAUJ={},  # pymatgen sets these, but they're all default values anyways
            LDAUL={
                "F": {
                    "Ag": 2,
                    "Co": 2,
                    "Cr": 2,
                    "Cu": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Nb": 2,
                    "Ni": 2,
                    "Re": 2,
                    "Ta": 2,
                    "V": 2,
                    "W": 2,
                },
                "O": {
                    "Ag": 2,
                    "Co": 2,
                    "Cr": 2,
                    "Cu": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Nb": 2,
                    "Ni": 2,
                    "Re": 2,
                    "Ta": 2,
                    "V": 2,
                    "W": 2,
                },
                "S": {
                    "Fe": 2,
                    "Mn": 2.5,
                },
            },
            LDAUU={
                "F": {
                    "Ag": 1.5,
                    "Co": 3.4,
                    "Cr": 3.5,
                    "Cu": 4,
                    "Fe": 4.0,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Nb": 1.5,
                    "Ni": 6,
                    "Re": 2,
                    "Ta": 2,
                    "V": 3.1,
                    "W": 4.0,
                },
                "O": {
                    "Ag": 1.5,
                    "Co": 3.4,
                    "Cr": 3.5,
                    "Cu": 4,
                    "Fe": 4.0,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Nb": 1.5,
                    "Ni": 6,
                    "Re": 2,
                    "Ta": 2,
                    "V": 3.1,
                    "W": 4.0,
                },
                "S": {
                    "Fe": 1.9,
                    "Mn": 2.5,
                },
            },
        ),
    )

    # These are some default error handlers to use
    error_handlers = [
        TetrahedronMesh(),
        RotationMatrix(),
        Brmix(),
        Zpotrf(),
        SubspaceMatrix(),
        IncorrectShift(),
        RealOptlay(),
        Tetirr(),
        RotationNonIntMatrix(),
        LongVector(),
        TripleProduct(),
        Pricel(),
        Brions(),
        Zbrent(),
        InsufficientBands(),
        Pssyevx(),
        Eddrmm(),
        Edddav(),
        Edwav(),
        Zheev(),
        ElfKpar(),
        Rhosyg(),
        Posmap(),
        PointGroup(),
        SymprecNoise(),
        IncorrectSmearing(),
        MeshSymmetry(),
        Unconverged(),
        NonConverging(),
        Potim(),
        PositiveEnergy(),
        Frozen(),
        LargeSigma(),
    ]



# -*- coding: utf-8 -*-

from simmate.workflow_engine import s3task_to_workflow
from simmate.calculators.vasp.tasks.population_analysis import (
    MatProjELF as MPPreBaderTask,
)
from simmate.calculators.vasp.database.population_analysis import (
    MatProjELF as MPELFResults,
)


workflow = s3task_to_workflow(
    name="population-analysis/elf-matproj",
    module=__name__,
    project_name="Simmate-PopulationAnalysis",
    s3task=MPPreBaderTask,
    calculation_table=MPELFResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)



# -*- coding: utf-8 -*-

from simmate.workflow_engine.workflow import (
    Workflow,
    Parameter,
    ModuleStorage,
)
from simmate.workflow_engine.common_tasks import LoadInputAndRegister, SaveOutputTask
from simmate.calculators.vasp.tasks.dynamics import MITDynamics
from simmate.calculators.vasp.database.dynamics import MITDynamicsRun

WORKFLOW_NAME = "dynamics/mit"

s3task_obj = MITDynamics()
load_input_and_register = LoadInputAndRegister(MITDynamicsRun)
load_input_and_register = LoadInputAndRegister(
    workflow_name=WORKFLOW_NAME,
    input_obj_name="structure",
    calculation_table=MITDynamicsRun,
)
save_results = SaveOutputTask(MITDynamicsRun)

with Workflow(WORKFLOW_NAME) as workflow:
    structure = Parameter("structure")
    command = Parameter("command", default="vasp_std > vasp.out")
    source = Parameter("source", default=None)
    directory = Parameter("directory", default=None)
    use_previous_directory = Parameter("use_previous_directory", default=False)

    # extra parameters unique to molecular dynamics runs
    temperature_start = Parameter("temperature_start", default=300)
    temperature_end = Parameter("temperature_end", default=1200)
    time_step = Parameter("time_step", default=2)
    nsteps = Parameter("nsteps", default=10000)

    structure_toolkit, directory_cleaned = load_input_and_register(
        input_obj=structure,
        source=source,
        directory=directory,
        command=command,
        use_previous_directory=use_previous_directory,
        temperature_start=temperature_start,
        temperature_end=temperature_end,
        time_step=time_step,
        nsteps=nsteps,
    )
    output = s3task_obj(
        structure=structure_toolkit,
        command=command,
        directory=directory_cleaned,
        temperature_start=temperature_start,
        temperature_end=temperature_end,
        time_step=time_step,
        nsteps=nsteps,
    )
    calculation_id = save_results(output=output)

workflow.storage = ModuleStorage(__name__)
workflow.project_name = "Simmate-Dynamics"
workflow.calculation_table = MITDynamicsRun
workflow.result_table = MITDynamicsRun
workflow.register_kwargs = (["prefect_flow_run_id", "structure", "source"],)
workflow.result_task = output
workflow.s3task = MITDynamics

# by default we just copy the docstring of the S3task to the workflow
workflow.__doc__ = MITDynamics.__doc__



# -*- coding: utf-8 -*-

from .load_input_and_register import LoadInputAndRegister
from .save_result import SaveOutputTask
from .load_nested_calculation import LoadNestedCalculationTask
from .save_nested_calculation import SaveNestedCalculationTask
from .parse_multi_command import parse_multi_command
from .customized_s3task import run_customized_s3task



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate database. All commands are 
accessible through the "simmate database" command.
"""

import click


@click.group()
def database():
    """A group of commands for managing your database."""
    pass


@database.command()
@click.option(
    "--confirm-delete",
    is_flag=True,
    help="automatically confirms that you want to delete your existing database",
)
def reset(confirm_delete):
    """Removes any existing data and sets up a clean database."""

    # make sure the user knows what they are doing and actually wants to continue
    if not confirm_delete:
        click.confirm(
            "Note that this deletes your current database and cannot be undone. "
            "We highly recommend you make a copy of your database before doing this. "
            "Do you still want to continue?",
            abort=True,
        )
    # We can now proceed with reseting the database
    click.echo("Removing database and rebuilding...")

    from simmate.configuration.django.database import reset_database

    reset_database()

    # Let the user know everything succeeded
    click.echo("Success! Your database has been reset.")


@database.command()
def update():
    """Updates the database with any changes made."""

    # check Django if there are any updates to be made
    click.echo("Checking for updates...")

    from simmate.configuration.django.database import update_database

    update_database()

    # Let the user know everything succeeded
    click.echo("Success! Your database tables are now up to date.")


@database.command()
@click.option(
    "--filename",
    "-f",
    default="database_dump.json",
    help="the file to write the all the JSON data to",
)
def dumpdata(filename):
    """Takes the Simmate database and writes it to a json file."""

    # Begin writing the database to the json file.
    click.echo("Writing all data to JSON...")

    from simmate.configuration.django.database import dump_database_to_json

    dump_database_to_json(filename=filename)

    # Let the user know everything succeeded
    click.echo(
        f"Success! You should now see the file {filename} with all of your data."
    )


@database.command()
@click.option(
    "--filename",
    "-f",
    default="database_dump.json",
    help="the file to load the all the JSON data from",
)
def loaddata(filename):
    """Takes a JSON database and loads it into the Simmate database."""

    # Begin writing the database to the json file.
    click.echo("Loading all data from JSON...")

    from simmate.configuration.django.database import load_database_from_json

    load_database_from_json(filename=filename)

    # Let the user know everything succeeded
    click.echo(
        f"Success! You now have all the data from {filename} available in your database."
    )


# explicitly list functions so that pdoc doesn't skip them
__all__ = ["database", "reset", "update", "dumpdata", "loaddata"]



# -*- coding: utf-8 -*-

from simmate.workflow_engine import s3task_to_workflow
from simmate.calculators.vasp.tasks.relaxation import (
    Quality01Relaxation as Quality01RelaxationTask,
)
from simmate.calculators.vasp.database.relaxation import (
    Quality01Relaxation as Quality01RelaxationResults,
)

workflow = s3task_to_workflow(
    name="relaxation/quality01",
    module=__name__,
    project_name="Simmate-Relaxation",
    s3task=Quality01RelaxationTask,
    calculation_table=Quality01RelaxationResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.population_analysis.pre_bader import (
    MatProjPreBader,
)


class MatProjPreBaderELF(MatProjPreBader):
    """
    Runs a static energy calculation with a high-density FFT grid under settings
    from the Materials Project. Results can be used for Bader analysis where
    the ELF is used as the reference instead of the CHGCAR.
    """

    incar = MatProjPreBader.incar.copy()
    incar.update(
        LELF=True,  # writes ELFCAR
        NPAR=1,  # must be set if LELF is set to True
        PREC="Single",  # ensures CHGCAR grid matches ELFCAR grid
    )



# -*- coding: utf-8 -*-

from .all_paths import workflow as all_paths_workflow
from .single_path import workflow as single_path_workflow
from .from_endpoints import workflow as from_endpoints_workflow
from .from_images import workflow as from_images_workflow



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.base import VaspTask
from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS,
)
from simmate.calculators.vasp.error_handlers import (
    TetrahedronMesh,
    RotationMatrix,
    Brmix,
    Zpotrf,
    SubspaceMatrix,
    IncorrectShift,
    RealOptlay,
    Tetirr,
    RotationNonIntMatrix,
    LongVector,
    TripleProduct,
    Pricel,
    Brions,
    Zbrent,
    InsufficientBands,
    Pssyevx,
    Eddrmm,
    Edddav,
    Edwav,
    Zheev,
    ElfKpar,
    Rhosyg,
    Posmap,
    PointGroup,
    SymprecNoise,
    IncorrectSmearing,
    MeshSymmetry,
    Unconverged,
    NonConverging,
    Potim,
    PositiveEnergy,
    Frozen,
    LargeSigma,
)


class MatProjRelaxation(VaspTask):
    """
    Runs a VASP geometry optimization using Materials Project settings.

    This is currently the highest-quality preset of all of Simmate's relaxations.

    Materials Project settings are often considered the minimum-required
    quality for publication and is sufficient for most applications. If you are
    looking at one structure in detail (for electronic, vibrational, and other
    properties), you should still test for convergence using higher-quality
    settings.
    """

    # This uses the PBE functional with POTCARs that have lower electron counts
    # and convergence criteria.
    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS

    # These are all input settings for this task. Note a lot of settings
    # depend on the structure/composition being analyzed.
    incar = dict(
        # These settings are the same for all structures regardless of composition
        ALGO="Fast",
        EDIFF__per_atom=5.0e-05,
        ENCUT=520,
        IBRION=2,
        ISIF=3,
        ISMEAR=-5,
        ISPIN=2,
        KSPACING=0.4,  # !!! This is where we are different from pymatgen right now
        LASPH=True,
        LORBIT=11,
        LREAL="Auto",
        LWAVE=False,
        NELM=100,
        NSW=99,
        PREC="Accurate",
        SIGMA=0.05,
        # The magnetic moments are dependent on what the composition and oxidation
        # states are. Note our default of 0.6 is different from the VASP default too.
        MAGMOM__smart_magmom={
            "default": 0.6,
            "Ce": 5,
            "Ce3+": 1,
            "Co": 0.6,
            "Co3+": 0.6,
            "Co4+": 1,
            "Cr": 5,
            "Dy3+": 5,
            "Er3+": 3,
            "Eu": 10,
            "Eu2+": 7,
            "Eu3+": 6,
            "Fe": 5,
            "Gd3+": 7,
            "Ho3+": 4,
            "La3+": 0.6,
            "Lu3+": 0.6,
            "Mn": 5,
            "Mn3+": 4,
            "Mn4+": 3,
            "Mo": 5,
            "Nd3+": 3,
            "Ni": 5,
            "Pm3+": 4,
            "Pr3+": 2,
            "Sm3+": 5,
            "Tb3+": 6,
            "Tm3+": 2,
            "V": 5,
            "W": 5,
            "Yb3+": 1,
        },
        # We run LDA+U for certain compositions. This is a complex configuration
        # so be sure to read the "__smart_ldau" modifier for more information.
        # But as an example for how the mappings work...
        # {"F": {"Ag": 2}} means if the structure is a fluoride, then we set
        # the MAGMOM for Ag to 2. Any element that isn't listed defaults to 0.
        multiple_keywords__smart_ldau=dict(
            LDAU__auto=True,
            LDAUTYPE=2,
            LDAUPRINT=1,
            LMAXMIX__auto=True,
            LDAUJ={},  # pymatgen sets these, but they're all default values anyways
            LDAUL={
                "F": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
                "O": {
                    "Co": 2,
                    "Cr": 2,
                    "Fe": 2,
                    "Mn": 2,
                    "Mo": 2,
                    "Ni": 2,
                    "V": 2,
                    "W": 2,
                },
            },
            LDAUU={
                "F": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
                "O": {
                    "Co": 3.32,
                    "Cr": 3.7,
                    "Fe": 5.3,
                    "Mn": 3.9,
                    "Mo": 4.38,
                    "Ni": 6.2,
                    "V": 3.25,
                    "W": 6.2,
                },
            },
        ),
    )

    # These are some default error handlers to use
    error_handlers = [
        TetrahedronMesh(),
        RotationMatrix(),
        Brmix(),
        Zpotrf(),
        SubspaceMatrix(),
        IncorrectShift(),
        RealOptlay(),
        Tetirr(),
        RotationNonIntMatrix(),
        LongVector(),
        TripleProduct(),
        Pricel(),
        Brions(),
        Zbrent(),
        InsufficientBands(),
        Pssyevx(),
        Eddrmm(),
        Edddav(),
        Edwav(),
        Zheev(),
        ElfKpar(),
        Rhosyg(),
        Posmap(),
        PointGroup(),
        SymprecNoise(),
        IncorrectSmearing(),
        MeshSymmetry(),
        Unconverged(),
        NonConverging(),
        Potim(),
        PositiveEnergy(),
        Frozen(),
        LargeSigma(),
    ]



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate workflow engine. All commands are 
accessible through the "simmate workflow-engine" command.
"""

import click


@click.group()
def workflow_engine():
    """
    A group of commands for starting up Prefect Agents and Dask Clusters.
    There is also a simple "Simmate Cluster" that you can use too, but it
    is only meant for quick testing.

    Setting up your computational resources can be tricky, so be sure to go
    through our tutorial before trying this on your own: <TODO: INSERT LINK>
    """
    pass


@workflow_engine.command()
def setup_cloud():
    """
    This configures Prefect Cloud for all of the Simmate Workflows. This includes
    registering them as well as organizing them into projects.

    If at any point you'd like to delete all of these and reset them, use the
    delete-cloud command followed by running this again.
    """
    click.echo("Building Prefect Projects and registering Simmate Workflows.")
    from simmate.configuration.prefect.projects import build

    build()


@workflow_engine.command()
def delete_cloud():
    """
    This clears your Prefect Cloud for all of the Simmate Projects and Workflows.

    To rebuild them, use the setup-cloud command. Note, you may have to wait
    a few minutes before running this second command because it takes a little
    for your deletion to take place in Prefect Cloud.
    """
    click.echo("Deleting all Prefect Projects and Simmate Workflows.")
    from simmate.configuration.prefect.projects import delete

    delete()


@workflow_engine.command()
@click.option(
    "--nitems_max",
    "-n",
    default=None,
    help="the number of jobs to run before shutdown",
)
@click.option(
    "--timeout",
    "-t",
    default=None,
    help="the time (in seconds) after which this worker will stop running jobs and shutdown",
)
@click.option(
    "--close_on_empty_queue",
    "-c",
    default=False,
    help="whether to shutdown or not when the queue is empty",
)
@click.option(
    "--waittime_on_empty_queue",
    "-w",
    default=60,
    help="how long to wait before checking again when the queue is found to be empty",
)
def start_worker(nitems_max, timeout, close_on_empty_queue, waittime_on_empty_queue):
    """
    This starts a Simmate Worker which will query the database for jobs to run.

    """

    from simmate.workflow_engine.execution.worker import SimmateWorker

    worker = SimmateWorker(
        nitems_max=nitems_max,
        timeout=timeout,
        close_on_empty_queue=close_on_empty_queue,
        waittime_on_empty_queue=waittime_on_empty_queue,
    )
    worker.start()


@workflow_engine.command()
def start_singletask_worker():
    """
    This starts a Simmate Worker that only runs one job and then shuts down. Also,
    if no job is available in the queue, it will shut down.

    Note: this can be acheived using the start-worker command too, but because
    this is such a common use-case, we include this command for convienence.
    It is the same as...

    simmate start-worker -n 1 -c True

    """

    from simmate.workflow_engine.execution.worker import SimmateWorker

    worker = SimmateWorker(
        nitems_max=1,
        timeout=None,
        close_on_empty_queue=True,
        waittime_on_empty_queue=0.1,
    )
    worker.start()


@workflow_engine.command()
@click.option(
    "--agent_name",
    "-n",
    help="the name of the agent that will be visible in prefect cloud",
)
@click.option(
    "--agent_labels",
    "-l",
    help="labels that the agent will query for. To list multiple do... -l label1 -l label2",
    multiple=True,
)
@click.option(
    "--n_workers",
    "-w",
    help="the number of separate workers to run/submit",
    type=int,
)
@click.option(
    "--cpus_per_worker",
    "-c",
    help="the number of cpus that each worker will request",
    type=int,
)
@click.option(
    "--memory_per_worker",
    "-m",
    help="the amount of memory that each worker will request",
)
@click.option(
    "--walltime_per_worker",
    "-t",
    help="the timelimit set for each worker",
)
def run_cluster(
    agent_name,
    agent_labels,
    n_workers,
    cpus_per_worker,
    memory_per_worker,
    walltime_per_worker,
):
    """
    This starts up a Dask cluster and a Prefect Agent in order to run Simmate jobs.

    This convenience command is really only meant for basic purposes, as Dask
    and Prefect teams offer their own commands with much more control. For
    example, this command uses Prefect's LocalAgent, but there are other
    advanced types available such as DockerAgent which may be better for your
    use case.

    If you would like this cluster to run endlessly in the background, you can
    submit it with something like "nohup simmate workflow-engine run-cluster &".
    The "nohup" and "&" symbol together make it so this runs in the background AND
    it won't shutdown if you close your terminal (or ssh). To stop this from running,
    you'll now need to find the running process and kill it. Use something like
    "ps -aef | grep simmate" to find the running process and grab its ID. Then
    kill that process id with "kill 123" (if the id was 123).
    """

    # All input arguments are optional. Therefore, I go through all of them and
    # remove the ones that weren't set. This prevents overwritting default settings
    # at a lower level. Note, I also rename several parameters (e.g. "cpus_per_worker").
    # These different names for the CLI only exist to make things easier for beginners.
    agent_kwargs = locals()  # grabs all input parameters as a dict
    possible_kwargs = list(agent_kwargs.keys())  # because we will be deleting keys
    for key in possible_kwargs:
        if not agent_kwargs[key]:
            agent_kwargs.pop(key)
            continue
        # renaming several parameters...
        if key == "cpus_per_worker":
            agent_kwargs["job_cpu"] = agent_kwargs.pop(key)
        elif key == "memory_per_worker":
            agent_kwargs["job_mem"] = agent_kwargs.pop(key)
        elif key == "walltime_per_job":
            agent_kwargs["walltime"] = agent_kwargs.pop(key)

    from simmate.configuration.prefect.setup_resources import run_cluster_and_agent

    run_cluster_and_agent(**agent_kwargs)


# explicitly list functions so that pdoc doesn't skip them
__all__ = [
    "workflow_engine",
    "setup_cloud",
    "delete_cloud",
    "start_worker",
    "run_cluster",
]



# -*- coding: utf-8 -*-

from simmate.workflow_engine import (
    s3task_to_workflow,
    task,
    Parameter,
    Workflow,
    ModuleStorage,
)
from simmate.calculators.vasp.tasks.population_analysis import (
    MatProjPreBader as MPPreBaderTask,
)
from simmate.calculators.vasp.database.population_analysis import (
    MatProjBaderAnalysis as MPBaderResults,
)
from simmate.calculators.bader.tasks import BaderAnalysis as BaderAnalysisTask

prebader_workflow = s3task_to_workflow(
    name="population-analysis/prebader-matproj",
    module=__name__,
    project_name="Simmate-PopulationAnalysis",
    s3task=MPPreBaderTask,
    calculation_table=MPBaderResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)

prebader_task = prebader_workflow.to_workflow_task()


@task
def save_bader_results(bader_result, prefect_flow_run_id):
    # load the results. We are particullary after the first result with
    # is a pandas dataframe of oxidation states.
    oxidation_data, extra_data = bader_result["result"]

    # load the calculation entry for this workflow run. This should already
    # exist thanks to the load_input_and_register task of the prebader workflow
    calculation = MPBaderResults.from_prefect_id(
        prefect_flow_run_id,
    )
    # BUG: can't use context to grab the id because workflow tasks generate a
    # different id than the main workflow

    # now update the calculation entry with our results
    calculation.oxidation_states = list(oxidation_data.oxidation_state.values)
    calculation.save()


bader_task = BaderAnalysisTask()


with Workflow("population-analysis/bader-matproj") as workflow:

    structure = Parameter("structure")
    command = Parameter("command", default="vasp_std > vasp.out")
    source = Parameter("source", default=None)
    directory = Parameter("directory", default=None)

    prebader_result = prebader_task(
        structure=structure,
        command=command,
        source=source,
        directory=directory,
    )

    # Bader only adds files and doesn't overwrite any, so I just run it
    # in the original directory. I may switch to copying over to a new
    # directory in the future though.
    bader_result = bader_task(directory=prebader_result["directory"])

    save_bader_results(bader_result, prebader_result["prefect_flow_run_id"])


workflow.storage = ModuleStorage(__name__)
workflow.project_name = "Simmate-PopulationAnalysis"
workflow.calculation_table = MPBaderResults
workflow.result_table = MPBaderResults
workflow.register_kwargs = ["prefect_flow_run_id", "structure", "source"]
workflow.result_task = bader_result
workflow.s3tasks = [MPPreBaderTask, BaderAnalysisTask]

workflow.__doc__ = """
    Runs a static energy calculation using an extra-fine FFT grid and then
    carries out Bader analysis on the resulting charge density.
"""



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure


class AflowPrototype(Structure):
    """
    A collection of prototype crystal structures from the AFLOW library.

    While this is still a table of structures, this should instead be viewed
    as a table of structure-types. All of the entries in this table come
    from the AFLOW Encyclopedia of Crystallographic Prototypes:
      http://www.aflowlib.org/prototype-encyclopedia/
    """

    class Meta:
        app_label = "prototypes"

    base_info = [
        "structure_string",
        "mineral_name",
        "aflow_id",
        "pearson_symbol",
        "strukturbericht",
        "nsites_wyckoff",
    ]
    source = "AFLOW Encyclopedia of Crystallographic Prototypes"
    source_doi = "https://doi.org/10.1016/j.commatsci.2017.01.017"
    # remote_archive_link = "TODO"  # also this will be private due to AFLOW licensing

    mineral_name = table_column.CharField(max_length=75, blank=True, null=True)
    """
    Common mineral name for this prototype (e.g. "rocksalt"). Note, not all 
    prototypes have this available.
    """

    aflow_id = table_column.CharField(max_length=30)
    """
    Identifier used by the AFLOW library. (e.g. "AB_hP6_154_a_b")
    """

    pearson_symbol = table_column.CharField(max_length=6)
    """
    Pearson symbol for the prototype structure
    """

    strukturbericht = table_column.CharField(max_length=6)
    """
    Strukturbericht symbol for the prototype structure
    """

    nsites_wyckoff = table_column.IntegerField()
    """
    Number of symmetrically unique wyckoff sites in the prototype structure
    """

    @property
    def name(self):
        """
        This helps piece together the name of the prototype in a user-friendly
        format. We start by checking if there is a mineral_name to use, and we
        also use the prototype's composition.

        An example of a structure with a mineral name is...
            Cinnabar (HgS) Structure-type
        And an example of a structure without a mineral name is..
            CaC6 Structure-type
        """
        if self.mineral_name:
            return f"{self.mineral_name} ({self.formula_reduced}) Structure-type"
        else:
            return f"{self.formula_reduced} Structure-type"



# -*- coding: utf-8 -*-

"""
This file establishes the info needed to install simmate via pip and also how to
upload it to PyPI. This file was copied and editted from the following source:
    https://github.com/pypa/sampleproject

For the details on conda-forge installation see:
    https://github.com/conda-forge/simmate-feedstock
"""

import os
from glob import glob
from setuptools import setup, find_packages
import pathlib

# Grab the full path to this file
here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
# long_description = (here / "README.md").read_text(encoding="utf-8")
# NOTE: I remove the long description and just have users move directly to github
long_description = (
    "Please do NOT install Simmate with pip. We recommend conda instead."
    " Visit our [github page](https://github.com/jacksund/simmate)"
    " for more information."
)


def get_package_data(base_dir):
    """
    This function goes through the src/simmate directory and grabs all of our
    non-python files (md, rst, json, csv, yaml, html). It returns a list of
    paths to these files with the relative path starting from src/simmate.

    These should all be passed to package_data to ensure these files are
    included when simmate is installed.
    """
    # convert PosixPath to string
    base_dir = str(base_dir)

    # all designate the path to source code
    source_dir = os.path.join("src", "simmate")

    queries = [
        "**/*.md",
        "**/*.rst",
        "**/*.json",
        "**/*.csv",
        "**/*.yaml",
        "**/*.html",
        "**/*.svg",
    ]
    # add the base directory to the start of each query
    # we also add the source directory
    queries = [os.path.join(base_dir, source_dir, q) for q in queries]

    # now grab all the files
    all_files = []
    for query in queries:
        files = glob(query, recursive=True)
        all_files += files

    # each filename will still start with base_dir, which we want to remove.
    all_files = [os.path.relpath(f, start=source_dir) for f in all_files]

    return all_files


# For debugging get_package_data
# raise Exception(get_package_data(here))

setup(
    # published name for pip install to use
    name="simmate",
    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    version="0.3.0",
    # a quick summary and then README
    description="Simmate is a toolbox for computational materials research.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # Link to our homepage. Use github for now.
    url="https://github.com/jacksund/simmate",
    # Lead dev info
    author="Jack D. Sundberg",
    author_email="jacksund@live.unc.edu",
    # Classifiers help users find your project by categorizing it.
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
    ],
    # Keywords -- Removed for now. Below are PyMatgen's keywors for future reference
    # keywords=(
    #     "VASP, gaussian, ABINIT, nwchem, qchem, materials, science, project, "
    #     "electronic, structure, analysis, phase, diagrams, crystal"
    # ),
    # Indicate which directory the source coude is in
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    # Specify which Python versions supported.
    python_requires=">=3.10",
    # This field lists other packages that your project depends on to run.
    install_requires=[
        # Core dependencies
        "numpy>=1.22.0",
        "pandas>=1.3.5",
        "django>=4.0.0",
        "prefect>=0.15.11",
        "dask>=2021.12.0",
        "click>=8.0.3",
        # Extra (smaller) dependencies & utilities
        "django-crispy-forms>=1.13.0",  # for formatting of online forms
        "django-pandas>=0.6.6",  # for converting QuerySets to PandasDataFrames
        "dask-jobqueue>=0.7.3",  # for submitting on clusters
        "dj-database-url>=0.5.0",  # for DigitalOcean URL conversion
        "djangorestframework>=3.13.1",  # for our REST API
        "django-filter>=21.1",  # sets up automatic filters for our REST API
        "pyyaml>=6.0",  # for yaml configuration files
        "plotly>=5.4.0",  # for interactive plots and visualization
        "tqdm>=4.62.3",  # for monitoring progress of long for-loops
        "pdoc>=11.0.0",  # for docs and markdown rendering in html templates
        #
        # For development and testing
        "pytest>=6.2.5",
        "pytest-django>=4.5.2",
        "pytest-mock>=3.7.0",
        "pytest-xdist>=2.5.0",
        "black>=22.1.0",
        "coverage>=6.2",
        #
        # These are from the MP stack and I want to phase them out over time
        "pymatgen>=2022.1.9",
        "pymatgen-analysis-diffusion>=2021.4.29",  # pymatgen-diffusion on conda
        "matminer>=0.7.6",
        #
        # These are packages that I commonly use alongside simmate. I plan to
        # organize these into optional dependencies and/or documentation. But until
        # then, I keep them here for my own reference.
        # "numba>=0.53.0",  # for speed-up of basic calcs
        # "psycopg2-binary>=2.9.2",  # for Postgres connections (added -binary to fix bug)
        # 'selenium',  # for web scraping (slow but robust)
        # 'spyder',  # IDE for writing/editting
        # 'gunicorn',  # for website server (Django+DigitalOcean) # !!! NOT WINDOWS
        # "graphviz==1.7",  # python-graphviz on conda. for viewing prefect flows
        # "pygraphviz==0.19",  # pygraphviz on conda. for viewing django tables
        # "scikit-learn>=1.0.1",  # for machine-learning
        # "fabric>=2.6.0",  # for remote ssh connections
        # "django-extensions>=3.1.5",  # simple tools to help with django development
    ],
    # Register command line interface
    entry_points={
        "console_scripts": [
            "simmate = simmate.command_line.base_command:simmate",
        ],
    },
    # All files that aren't *.py need to be defined explicitly. Don't "automate"
    # this to grab all files because this could break installation. This can
    # be effectively the opposite of .gitignore.
    include_package_data=True,
    package_data={
        # Recursive calls are not supported yet, so keep an eye on this issue
        # https://github.com/pypa/setuptools/issues/1806
        # "simmate": [
        #     "**/*.md",
        #     "**/*.rst",
        #     "**/*.json",
        #     "**/*.csv",
        #     "**/*.yaml",
        #     "**/*.html",
        # ],
        # For now, I make a custom function that calls glob recursively
        "simmate": get_package_data(here),
    },
)



# -*- coding: utf-8 -*-

import os
import yaml
from pandas import DataFrame

from pymatgen.io.vasp.outputs import Chgcar
from pymatgen.io.vasp import Potcar

from simmate.toolkit import Structure
from simmate.workflow_engine import S3Task
from simmate.calculators.bader.outputs import ACF


# TODO: The chgsum.pl script will be replaced with a simple python function
# that just sums the two files. It might not be as fast but it removes one
# executable file from having to be in the user's path. So in the future, this
# Task will be depreciated/removed.
class CombineCHGCARs(S3Task):
    """
    This tasks simply sums two charge density files into a new file. It uses
    a script from the Henkleman group.
    """

    command = "chgsum.pl AECCAR0 AECCAR2 > chgsum.out"


class BaderAnalysis(S3Task):

    command = "bader CHGCAR -ref CHGCAR_sum -b weight > bader.out"
    """
    The command to call the executable, which is typically bader. Note we
    use the `-b weight` by default, which means we apply the weight method for
    partitioning from of 
    [Yu and Trinkle](http://theory.cm.utexas.edu/henkelman/code/bader/download/yu11_064111.pdf).
    """

    def setup(self, structure: Structure, directory: str):
        """
        Bader analysis requires that a static-energy calculation be ran beforehand
        - typically using VASP. This setup therefore just involves ensuring that
        the proper files are present.
        """

        # Make sure that there are the proper output files from a VASP calc
        files = ["CHGCAR", "AECCAR0", "AECCAR2", "POTCAR"]
        filenames = [os.path.join(directory, file) for file in files]
        if not all(os.path.exists(filename) for filename in filenames):
            raise Exception(
                "A static energy calculation is required before running Bader "
                "analysis. The following files must exist in the directory where "
                f"this task is ran: {files}"
            )

        # Make the CHGCAR_sum file using Bader's helper script
        CombineCHGCARs().run(directory=directory)

    def workup(self, directory: str):
        """
        A basic workup process that reads Bader analysis results from the ACF.dat
        file and calculates the corresponding oxidation states with the existing
        POTCAR files.
        """

        # load the ACF.dat file
        acf_filename = os.path.join(directory, "ACF.dat")
        dataframe, extra_data = ACF(filename=acf_filename)

        # load the electron counts used by VASP from the POTCAR files
        # OPTIMIZE: this can be much faster if I have a reference file
        potcar_filename = os.path.join(directory, "POTCAR")
        potcars = Potcar.from_file(potcar_filename)
        nelectron_data = {}
        # the result is a list because there can be multiple element potcars
        # in the file (e.g. for NaCl, POTCAR = POTCAR_Na + POTCAR_Cl)
        for potcar in potcars:
            nelectron_data.update({potcar.element: potcar.nelectrons})

        # grab the structure from the CHGCAR
        # OPTIMIZE: I should just grab from the POSCAR or CONTCAR for speed.
        # The reason I don't at the moment is because there may be empty atoms.
        chgcar_filename = os.path.join(directory, "CHGCAR")
        chgcar = Chgcar.from_file(chgcar_filename)
        structure = chgcar.structure

        # Calculate the oxidation state of each site where it is simply the
        # change in number of electrons associated with it from vasp potcar vs
        # the bader charge I also add the element strings for filtering functionality
        elements = []
        oxi_state_data = []
        for site, site_charge in zip(structure, dataframe.charge.values):
            element_str = site.specie.name
            elements.append(element_str)
            oxi_state = nelectron_data[element_str] - site_charge
            oxi_state_data.append(oxi_state)

        # add the new column to the dataframe
        dataframe = dataframe.assign(
            oxidation_state=oxi_state_data,
            element=elements,
        )
        # !!! There are multiple ways to do this, but I don't know which is best
        # dataframe["oxidation_state"] = pandas.Series(
        #     oxi_state_data, index=dataframe.index)

        # write output files/plots for the user to quickly reference
        self._write_output_summary(directory, dataframe, extra_data)

        # return all of our results
        return dataframe, extra_data

    def _write_output_summary(
        self, directory: str, dataframe: DataFrame, extra_data: dict
    ):
        """
        This prints a "simmate_summary.yaml" file with key output information.

        This method should not be called directly as it used within workup().
        """

        # write output of the dataframe
        summary_csv_filename = os.path.join(directory, "simmate_summary_bader.csv")
        dataframe.to_csv(summary_csv_filename)

        summary = {
            "notes": "view simmate_summary_bader.csv for more information",
            **extra_data,
        }

        summary_filename = os.path.join(directory, "simmate_summary.yaml")
        with open(summary_filename, "w") as file:
            content = yaml.dump(summary)
            file.write(content)



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine.workflow import (
    Workflow,
    Parameter,
    ModuleStorage,
)
from simmate.workflow_engine.common_tasks import (
    LoadInputAndRegister,
    parse_multi_command,
)

from simmate.calculators.vasp.workflows.nudged_elastic_band.utilities import (
    get_migration_images_from_endpoints,
)
from simmate.calculators.vasp.workflows.relaxation import (
    neb_endpoint_workflow as relaxation_neb_endpoint_workflow,
)
from simmate.calculators.vasp.workflows.nudged_elastic_band.from_images import (
    workflow as neb_from_images,
)
from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
)

WORKFLOW_NAME = "diffusion/from-endpoints"

# Convert our workflow objects to task objects
relax_endpoint = relaxation_neb_endpoint_workflow.to_workflow_task()
run_neb = neb_from_images.to_workflow_task()

# Extra setup tasks
load_input_and_register = LoadInputAndRegister(
    workflow_name=WORKFLOW_NAME,
    input_obj_name="structure",
)

with Workflow(WORKFLOW_NAME) as workflow:

    supercell_start = Parameter("supercell_start")
    supercell_end = Parameter("supercell_end")
    directory = Parameter("directory", default=None)
    source = Parameter("source", default=None)

    # This helps link to a higher-level table.
    diffusion_analysis_id = Parameter("diffusion_analysis_id", default=None)

    # I separate these out because each calculation is a very different scale.
    # For example, you may want to run the bulk relaxation on 10 cores, the
    # supercell on 50, and the NEB on 200. Even though more cores are available,
    # running smaller calculation on more cores could slow down the calc.
    command = Parameter("command", default="vasp_std > vasp.out")
    # command list expects three subcommands:
    #   command_bulk, command_supercell, and command_neb
    subcommands = parse_multi_command(
        command,
        commands_out=["command_supercell", "command_neb"],
    )

    # Load our input and make a base directory for all other workflows to run
    # within for us.
    # OPTIMIZE: is there a cleaner way to load two structures? Also,
    # directory_cleaned is grabbed twice here.
    supercell_start_toolkit, directory_cleaned = load_input_and_register(
        input_obj=supercell_start,
        source=source,
        directory=directory,
        command=command,
        diffusion_analysis_id=diffusion_analysis_id,
    )
    supercell_end_toolkit, directory_cleaned = load_input_and_register(
        input_obj=supercell_end,
        source=source,
        directory=directory,
        command=command,
        diffusion_analysis_id=diffusion_analysis_id,
    )
    # BUG: if directory isn't supplied, then we'd be making two directories
    # here -- and then only using the 2nd.
    # BUG: simmate_metadata.yaml will not include the supercell_start and
    # supercell_end it.

    # Relax the starting supercell structure
    run_id_00 = relax_endpoint(
        structure=supercell_start_toolkit,
        command=subcommands["command_supercell"],
        directory=directory_cleaned + os.path.sep + "endpoint_relaxation_start",
    )

    # Relax the ending supercell structure
    run_id_01 = relax_endpoint(
        structure=supercell_end_toolkit,
        command=subcommands["command_supercell"],
        directory=directory_cleaned + os.path.sep + "endpoint_relaxation_end",
    )

    images = get_migration_images_from_endpoints(
        supercell_start={
            "calculation_table": "NEBEndpointRelaxation",
            "directory": run_id_00["directory"],
            "structure_field": "structure_final",
        },
        supercell_end={
            "calculation_table": "NEBEndpointRelaxation",
            "directory": run_id_01["directory"],
            "structure_field": "structure_final",
        },
    )

    # Run NEB on this set of images
    run_id_02 = run_neb(
        migration_images=images,
        command=subcommands["command_neb"],
        source=source,
        directory=directory_cleaned,
        diffusion_analysis_id=diffusion_analysis_id,
    )


workflow.storage = ModuleStorage(__name__)
workflow.project_name = "Simmate-Diffusion"
# workflow.calculation_table = MITDiffusionAnalysis  # not implemented yet
# workflow.register_kwargs = ["prefect_flow_run_id"]
workflow.result_table = MITDiffusionAnalysis
workflow.s3tasks = [
    relaxation_neb_endpoint_workflow.s3task,
    neb_from_images.s3task,
]
workflow.__doc__ = """
    Runs a full diffusion analysis from a start and end supercell using NEB.
    
    The input start/end structures will be relaxed using the relaxation/neb-endpoint 
    workflow and then interpolated to generate 7 images. These 7 images are
    then relaxed using NEB within the diffusion/from-images workflow.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/neb-endpoint (for both start and end supercells)
        - a mini task that interpolated relaxed endpoints and makes images
        - diffusion/from-images
    
    If you are running this workflow via the command-line, you can run this 
    with...
    
    ``` bash
    simmate workflows run diffusion/all-paths -c "cmd1; cmd2" --supercell_start example1.cif --supercell_end example2.cif
    ```
    
    Note, the `-c` here is very important! Here we are passing three commands
    separated by semicolons. Each command is passed to a specific workflow call:
        
        - cmd1 --> used for endpoint supercell relaxations
        - cmd2 --> used for NEB
    
    Thus, you can scale your resources for each step. Here's a full -c option:
    
    ```
    -c "mpirun -n 12 vasp_std > vasp.out; mpirun -n 70 vasp_std > vasp.out"
    ```
"""



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Pricel(ErrorHandler):
    """
    This fixes an internal VASP error by turning off symmetry.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["internal error in subroutine PRICEL"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # make the fix
        incar["SYMPREC"] = 1e-8
        incar["ISYM"] = 0
        correction = "switched ISYM to 0 and SYMPREC to 1e-8"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

import itertools

from simmate.configuration.django import setup_full  # sets database connection
from simmate.database.prototypes.aflow import AflowPrototype

# from simmate.utilities import get_sanitized_structure, estimate_volume

from django_pandas.io import read_frame


class ExactAflowPrototypeStructures:

    """
    Given a compositon, this class can be used to create structures from
    the **exactly-matching** prototypes in the AFLOW library. This means we
    are only using prototypes that have the same anonymous formula. For example,
    if you request structures for Ca2N to be made, only prototypes that have
    the AB2 anonymoous formula will be used (such as the MoS2 prototype).
    """

    def __init__(self, composition, max_sites=115):
        # composition --> pymatgen composition object (ideally a reduced composition)
        # allow_multiples --> whether to allow structures where the reduced
        # compositions are the same, but the actual number of sites differs.
        # For example, Ca6N3 would be allow to return structures that give
        # Ca2N, Ca8N4, and so forth.
        # max_sites defaults to 115 because this would include all prototypes
        # in the aflow database (current largest is (105 sites))

        # Store these input properties for reference
        self.composition = composition
        self.max_sites = max_sites

        # ---------------------------------------------------------------------

        # First let's load all of the prototype structures that have a compatible
        # number of sites in the structure. For this we require that the prototypes
        # meet the following:
        #   (1) prototypes have the same anonymous formula as target the composition
        #   (2) the number of wykcoff sites is at least the number of target elements
        # Note: #1 ensures that #2 is true here, so we don't actually need to
        # check for both -- just #1.

        # We allow for multiples so we want all multiples nsites in our
        # desired composition. For example if our compositon, has 6 sites, we
        # would pull all prototypes that have 6, 12, 19, ..., 6*N sites.
        # First generate a list of the allow number of sites
        target_nsites_list = [
            int(composition.num_atoms * n)
            for n in range(1, (max_sites // int(composition.num_atoms)) + 1)
        ]
        # Now filter off the prototypes that have the proper number of sites
        prototype_entries = AflowPrototype.objects.filter(
            formula_anonymous=composition.anonymized_formula,
            nsites__in=target_nsites_list,
        ).all()

        # for user convience, let's store these selected prototypes as a dataframe
        # and attach it to this class instance
        self.prototype_dataframe = read_frame(prototype_entries)

        # Now that we have the prototype entries, let's convert them to
        # structure objects for us to use below.
        self.prototype_structures = [
            prototype.to_toolkit() for prototype in prototype_entries
        ]

        # ---------------------------------------------------------------------

        # We now generate all prototypes here. We store them in a final list
        # as we make them too.
        self.structures = []
        for prototype in self.prototype_structures:

            # we need to match elements in our target composition to those in
            # the prototype. For example, Ca2NF with the NaCrS2 prototype would
            # need to match S-->Ca, Cr-->N, and Na-->F.
            #
            # To do this, we need to look at every possible order of target
            # elements. We then iterate through these combinations and check
            # which ones are valid. For example, a Ca2NCl would give us possible
            # orderings of Ca-N-F, Ca-F-N, N-Ca-F, N-F-Ca, F-Ca-N, and F-N-Ca.
            # We then go through each ordering and see which ones will give our
            # target composition when substituted for Na-Cr-S in order.
            for target_elements in itertools.permutations(composition.elements):

                # Using this ordering of target elements we can now
                # attempt the subsitution for the prototype

                # We first sure to make a copy of the prototype that we can alter.
                new_structure = prototype.copy()

                # Go through and substitute out the elements
                for target_element, current_element in zip(
                    target_elements, prototype.composition
                ):
                    new_structure[current_element] = target_element

                # Now let's check to see that the final composition matches
                # what we are trying to produce. If so, we store the structure!
                # Note, we are after the reduce composition because we want
                # to allow things like Ca2NF == Ca4N2F2 as True.
                if (
                    new_structure.composition.reduced_composition
                    == composition.reduced_composition
                ):
                    self.structures.append(new_structure)

        # ---------------------------------------------------------------------

        # We now have all of the structures, but they will have unreasonable
        # volume and atomic distances right now. We iterate through each
        # and try to fix this by doing the following:
        #   (1) Scaling the structure to a predicted volume
        #   (2) Sanitizing the structure
        #   (3) [TODO] scaling the lattice vectors to avoid atoms being too close
        for i, structure in enumerate(self.structures):
            volume_guess = estimate_volume(structure.composition)
            structure.scale_lattice(volume_guess)
            structure_santized = get_sanitized_structure(structure)
            self.structures[i] = structure_santized



# -*- coding: utf-8 -*-

from simmate.workflow_engine import s3task_to_workflow
from simmate.calculators.vasp.tasks.density_of_states import (
    MatProjDensityOfStates as MatProjDensityOfStatesTask,
)
from simmate.calculators.vasp.database.density_of_states import (
    MatProjDensityOfStates as MatProjDensityOfStatesResults,
)

workflow = s3task_to_workflow(
    name="density-of-states/matproj",
    module=__name__,
    project_name="Simmate-Density-of-States",
    s3task=MatProjDensityOfStatesTask,
    calculation_table=MatProjDensityOfStatesResults,
    register_kwargs=["prefect_flow_run_id", "structure", "source"],
)
